{-# OPTIONS_GHC -fno-warn-orphans #-}
module Darcs.Patch.Prim
    ( primIsAddfile, primIsHunk, primIsBinary, primIsSetpref
    , primIsAdddir, is_filepatch
    , summarizePrim
    , applyPrimFL
    , PrimRead(..)
    , PrimShow(..)
    , FromPrim(..), FromPrims(..), ToFromPrim(..)
    , PrimPatch, PrimPatchBase(..)
    , PrimConstruct(..)
    , PrimCanonize(..)
    , PrimPatchCommon
    ) where

import Darcs.Patch.Prim.Class
    ( PrimConstruct(..), PrimCanonize(..)
    , PrimClassify(..), PrimDetails(..)
    , PrimShow(..), PrimRead(..)
    , PrimApply(..)
    , FromPrim(..), FromPrims(..), ToFromPrim(..)
    , PrimPatchBase(..), PrimPatch
    , PrimPatchCommon
    )
